@@ -1,4 +1,4 @@
-use crate::types::{Asn1Readable, Tlv};
+use crate::types::{Asn1Readable, SimpleAsn1Readable, Tlv};
 use crate::Tag;
 use core::fmt;
 
@@ -62,7 +62,7 @@ pub struct ParseError {
 }
 
 impl ParseError {
-    pub fn new(kind: ParseErrorKind) -> ParseError {
+    pub const fn new(kind: ParseErrorKind) -> ParseError {
         ParseError {
             kind,
             parse_locations: [None, None, None, None],
@@ -140,19 +140,14 @@ impl fmt::Display for ParseError {
             ParseErrorKind::InvalidSize { min, max, actual } => {
                 write!(
                     f,
-                    "invalid container size (expected between {} and {}, got {})",
-                    min, max, actual
+                    "invalid container size (expected between {min} and {max}, got {actual})"
                 )
             }
             ParseErrorKind::UnexpectedTag { actual } => {
-                write!(f, "unexpected tag (got {:?})", actual)
+                write!(f, "unexpected tag (got {actual:?})")
             }
             ParseErrorKind::ShortData { needed } => {
-                write!(
-                    f,
-                    "short data (needed at least {} additional bytes)",
-                    needed
-                )
+                write!(f, "short data (needed at least {needed} additional bytes)")
             }
             ParseErrorKind::IntegerOverflow => write!(f, "integer overflow"),
             ParseErrorKind::ExtraData => write!(f, "extra data"),
@@ -231,7 +226,8 @@ impl<'a> Parser<'a> {
         Parser::new(self.data)
     }
 
-    pub(crate) fn peek_tag(&mut self) -> Option<Tag> {
+    /// Returns the tag of the next element, without consuming it.
+    pub fn peek_tag(&mut self) -> Option<Tag> {
         let (tag, _) = Tag::from_bytes(self.data).ok()?;
         Some(tag)
     }
@@ -272,7 +268,7 @@ impl<'a> Parser<'a> {
             }
             0x82 => {
                 let length_bytes = self.read_bytes(2)?;
-                let length = usize::from(length_bytes[0]) << 8 | usize::from(length_bytes[1]);
+                let length = (usize::from(length_bytes[0]) << 8) | usize::from(length_bytes[1]);
                 // Enforce that we're not using long form for values <0x80,
                 // and that the first byte of the length is not zero (i.e.
                 // that we're minimally encoded)
@@ -283,8 +279,8 @@ impl<'a> Parser<'a> {
             }
             0x83 => {
                 let length_bytes = self.read_bytes(3)?;
-                let length = usize::from(length_bytes[0]) << 16
-                    | usize::from(length_bytes[1]) << 8
+                let length = (usize::from(length_bytes[0]) << 16)
+                    | (usize::from(length_bytes[1]) << 8)
                     | usize::from(length_bytes[2]);
                 // Same thing as the 0x82 case
                 if length < 0x10000 {
@@ -294,9 +290,9 @@ impl<'a> Parser<'a> {
             }
             0x84 => {
                 let length_bytes = self.read_bytes(4)?;
-                let length = usize::from(length_bytes[0]) << 24
-                    | usize::from(length_bytes[1]) << 16
-                    | usize::from(length_bytes[2]) << 8
+                let length = (usize::from(length_bytes[0]) << 24)
+                    | (usize::from(length_bytes[1]) << 16)
+                    | (usize::from(length_bytes[2]) << 8)
                     | usize::from(length_bytes[3]);
                 // Same thing as the 0x82 case
                 if length < 0x1000000 {
@@ -338,6 +334,32 @@ impl<'a> Parser<'a> {
     pub fn read_element<T: Asn1Readable<'a>>(&mut self) -> ParseResult<T> {
         T::parse(self)
     }
+
+    /// This is an alias for `read_element::<Explicit<T, tag>>` for use when
+    /// the tag is not known at compile time.
+    pub fn read_explicit_element<T: Asn1Readable<'a>>(&mut self, tag: u32) -> ParseResult<T> {
+        let expected_tag = crate::explicit_tag(tag);
+        let tlv = self.read_tlv()?;
+        if tlv.tag != expected_tag {
+            return Err(ParseError::new(ParseErrorKind::UnexpectedTag {
+                actual: tlv.tag,
+            }));
+        }
+        parse_single(tlv.data())
+    }
+
+    /// This is an alias for `read_element::<Implicit<T, tag>>` for use when
+    /// the tag is not known at compile time.
+    pub fn read_implicit_element<T: SimpleAsn1Readable<'a>>(&mut self, tag: u32) -> ParseResult<T> {
+        let expected_tag = crate::implicit_tag(tag, T::TAG);
+        let tlv = self.read_tlv()?;
+        if tlv.tag != expected_tag {
+            return Err(ParseError::new(ParseErrorKind::UnexpectedTag {
+                actual: tlv.tag,
+            }));
+        }
+        T::parse_data(tlv.data())
+    }
 }
 
 #[cfg(test)]
@@ -411,7 +433,7 @@ mod tests {
                 "ParseError { kind: InvalidValue, location: [\"Abc::123\", 12] }",
             ),
         ] {
-            assert_eq!(&format!("{:?}", e), expected);
+            assert_eq!(&format!("{e:?}"), expected);
         }
     }
 
@@ -473,7 +495,7 @@ mod tests {
             ),
         ]
         {
-            assert_eq!(&format!("{}", e), expected);
+            assert_eq!(&format!("{e}"), expected);
         }
     }
 
@@ -562,6 +584,15 @@ mod tests {
         assert_eq!(result, Err(ParseError::new(ParseErrorKind::ExtraData)));
     }
 
+    #[test]
+    fn test_peek_tag() {
+        let result = crate::parse(b"\x02\x01\x7f", |p| {
+            assert_eq!(p.peek_tag(), Some(Tag::primitive(0x02)));
+            p.read_element::<u8>()
+        });
+        assert_eq!(result, Ok(127));
+    }
+
     #[test]
     fn test_errors() {
         #[derive(Debug, PartialEq, Eq)]
@@ -1954,6 +1985,52 @@ mod tests {
                 b"",
             ),
         ]);
+        assert_parses_cb(
+            &[
+                (Ok(true), b"\x82\x01\xff"),
+                (Ok(false), b"\x82\x01\x00"),
+                (
+                    Err(ParseError::new(ParseErrorKind::ShortData { needed: 1 })),
+                    b"",
+                ),
+                (
+                    Err(ParseError::new(ParseErrorKind::UnexpectedTag {
+                        actual: Tag::primitive(0x01),
+                    })),
+                    b"\x01\x01\xff",
+                ),
+                (
+                    Err(ParseError::new(ParseErrorKind::UnexpectedTag {
+                        actual: Tag::primitive(0x02),
+                    })),
+                    b"\x02\x01\xff",
+                ),
+            ],
+            |p| p.read_implicit_element::<bool>(2),
+        );
+        assert_parses_cb(
+            &[
+                (Ok(Sequence::new(b"abc")), b"\xa2\x03abc"),
+                (Ok(Sequence::new(b"")), b"\xa2\x00"),
+                (
+                    Err(ParseError::new(ParseErrorKind::ShortData { needed: 1 })),
+                    b"",
+                ),
+                (
+                    Err(ParseError::new(ParseErrorKind::UnexpectedTag {
+                        actual: Tag::primitive(0x01),
+                    })),
+                    b"\x01\x01\xff",
+                ),
+                (
+                    Err(ParseError::new(ParseErrorKind::UnexpectedTag {
+                        actual: Tag::primitive(0x02),
+                    })),
+                    b"\x02\x01\xff",
+                ),
+            ],
+            |p| p.read_implicit_element::<Sequence<'_>>(2),
+        );
     }
 
     #[test]
@@ -1980,6 +2057,29 @@ mod tests {
                 b"\xa2\x03\x03\x01\xff",
             ),
         ]);
+        assert_parses_cb(
+            &[
+                (Ok(true), b"\xa2\x03\x01\x01\xff"),
+                (Ok(false), b"\xa2\x03\x01\x01\x00"),
+                (
+                    Err(ParseError::new(ParseErrorKind::ShortData { needed: 1 })),
+                    b"",
+                ),
+                (
+                    Err(ParseError::new(ParseErrorKind::UnexpectedTag {
+                        actual: Tag::primitive(0x01),
+                    })),
+                    b"\x01\x01\xff",
+                ),
+                (
+                    Err(ParseError::new(ParseErrorKind::UnexpectedTag {
+                        actual: Tag::primitive(0x03),
+                    })),
+                    b"\xa2\x03\x03\x01\xff",
+                ),
+            ],
+            |p| p.read_explicit_element::<bool>(2),
+        );
     }
 
     #[test]
