@@ -8,6 +8,9 @@ use core::borrow::Borrow;
 use core::hash::{Hash, Hasher};
 use core::marker::PhantomData;
 use core::mem;
+use core::num::{
+    NonZeroI16, NonZeroI32, NonZeroI64, NonZeroI8, NonZeroU16, NonZeroU32, NonZeroU64, NonZeroU8,
+};
 
 use crate::writer::Writer;
 use crate::{
@@ -17,14 +20,25 @@ use crate::{
 
 /// Any type that can be parsed as DER ASN.1.
 pub trait Asn1Readable<'a>: Sized {
+    /// Parse a value from the given parser.
+    ///
+    /// This method should read exactly one ASN.1 TLV from the parser,
+    /// consuming the appropriate bytes and returning the parsed value.
     fn parse(parser: &mut Parser<'a>) -> ParseResult<Self>;
+
+    /// Returns whether this type can parse values with the given tag.
     fn can_parse(tag: Tag) -> bool;
 }
 
 /// Types with a fixed-tag that can be parsed as DER ASN.1
 pub trait SimpleAsn1Readable<'a>: Sized {
+    /// The ASN.1 tag that this type expects when parsing.
     const TAG: Tag;
 
+    /// Parse the value from the given data bytes.
+    ///
+    /// This method receives the value portion of a TLV (without the tag or
+    /// length) and should parse it into the appropriate type.
     fn parse_data(data: &'a [u8]) -> ParseResult<Self>;
 }
 
@@ -56,29 +70,81 @@ impl<'a, T: SimpleAsn1Readable<'a>> SimpleAsn1Readable<'a> for Box<T> {
 
 /// Any type that can be written as DER ASN.1.
 pub trait Asn1Writable: Sized {
+    /// Write this value to the given writer.
+    ///
+    /// This method should write the complete ASN.1 encoding of this value,
+    /// including the tag, length, and content bytes.
     fn write(&self, dest: &mut Writer<'_>) -> WriteResult;
+
+    /// Get the complete encoded length (tag + length + content), if it can be
+    /// calculated efficiently.
+    ///
+    /// It is always safe to return `None`, which indicates the length is
+    /// unknown. Returning `Some(...)` from this method reduces the number of
+    /// re-allocations required in writing.
+    fn encoded_length(&self) -> Option<usize>;
 }
 
-// Types with a fixed-tag that can be written as DER ASN.1.
+/// Types with a fixed-tag that can be written as DER ASN.1.
 pub trait SimpleAsn1Writable: Sized {
+    /// The ASN.1 tag that this type uses when writing.
     const TAG: Tag;
 
+    /// Write the value's data to the given buffer.
+    ///
+    /// This method should write only the value bytes (without the tag and
+    /// length) to the buffer.
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult;
+
+    /// Get the length of the data content (without tag and length bytes) if it
+    /// can be calculated efficiently.
+    ///
+    /// It is always safe to return `None`, which indicates the length is
+    /// unknown. Returning `Some(...)` from this method reduces the number of
+    /// re-allocations required in writing.
+    fn data_length(&self) -> Option<usize>;
 }
 
+/// A trait for types that can be parsed based on a `DEFINED BY` value.
+///
+/// `T` is the type of the `DEFINED BY` field (nearly always `ObjectIdentifier`).
 pub trait Asn1DefinedByReadable<'a, T: Asn1Readable<'a>>: Sized {
+    /// Parse a value based on the previously parsed item.
+    ///
+    /// The `item` parameter contains the value that determines how to parse
+    /// the current value from the parser.
     fn parse(item: T, parser: &mut Parser<'a>) -> ParseResult<Self>;
 }
 
+/// A trait for types that can be written based on a `DEFINED BY` value.
+///
+/// `T` is the type of the `DEFINED BY` field (nearly always `ObjectIdentifier`).
 pub trait Asn1DefinedByWritable<T: Asn1Writable>: Sized {
+    /// Get a reference to the `DEFINED BY` value.
     fn item(&self) -> &T;
+
+    /// Write this value to the given writer.
     fn write(&self, dest: &mut Writer<'_>) -> WriteResult;
+
+    /// Get the complete encoded length (tag + length + content), if it can be
+    /// calculated efficiently.
+    ///
+    /// It is always safe to return `None`, which indicates the length is
+    /// unknown. Returning `Some(...)` from this method reduces the number of
+    /// re-allocations required in writing.
+    fn encoded_length(&self) -> Option<usize>;
 }
 
 impl<T: SimpleAsn1Writable> Asn1Writable for T {
     #[inline]
     fn write(&self, w: &mut Writer<'_>) -> WriteResult {
-        w.write_tlv(Self::TAG, move |dest| self.write_data(dest))
+        w.write_tlv(Self::TAG, self.data_length(), move |dest| {
+            self.write_data(dest)
+        })
+    }
+
+    fn encoded_length(&self) -> Option<usize> {
+        Some(Tlv::full_length(Self::TAG, self.data_length()?))
     }
 }
 
@@ -87,6 +153,10 @@ impl<T: SimpleAsn1Writable> SimpleAsn1Writable for &T {
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
         T::write_data(self, dest)
     }
+
+    fn data_length(&self) -> Option<usize> {
+        T::data_length(self)
+    }
 }
 
 impl<T: SimpleAsn1Writable> SimpleAsn1Writable for Box<T> {
@@ -94,6 +164,10 @@ impl<T: SimpleAsn1Writable> SimpleAsn1Writable for Box<T> {
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
         T::write_data(self, dest)
     }
+
+    fn data_length(&self) -> Option<usize> {
+        T::data_length(self)
+    }
 }
 
 /// A TLV (type, length, value) represented as the tag and bytes content.
@@ -109,6 +183,10 @@ pub struct Tlv<'a> {
 }
 
 impl<'a> Tlv<'a> {
+    pub(crate) fn full_length(t: Tag, inner_length: usize) -> usize {
+        t.encoded_length() + crate::writer::length_encoding_size(inner_length) + inner_length
+    }
+
     /// The tag portion of a TLV.
     pub fn tag(&self) -> Tag {
         self.tag
@@ -141,7 +219,13 @@ impl<'a> Asn1Readable<'a> for Tlv<'a> {
 impl Asn1Writable for Tlv<'_> {
     #[inline]
     fn write(&self, w: &mut Writer<'_>) -> WriteResult {
-        w.write_tlv(self.tag, move |dest| dest.push_slice(self.data))
+        w.write_tlv(self.tag, Some(self.data.len()), move |dest| {
+            dest.push_slice(self.data)
+        })
+    }
+
+    fn encoded_length(&self) -> Option<usize> {
+        Some(Tlv::full_length(self.tag, self.data.len()))
     }
 }
 
@@ -150,6 +234,10 @@ impl Asn1Writable for &Tlv<'_> {
     fn write(&self, w: &mut Writer<'_>) -> WriteResult {
         Tlv::write(self, w)
     }
+
+    fn encoded_length(&self) -> Option<usize> {
+        Tlv::encoded_length(self)
+    }
 }
 
 /// The ASN.1 NULL type, for use with `Parser.read_element` and
@@ -174,6 +262,10 @@ impl SimpleAsn1Writable for Null {
     fn write_data(&self, _dest: &mut WriteBuf) -> WriteResult {
         Ok(())
     }
+
+    fn data_length(&self) -> Option<usize> {
+        Some(0)
+    }
 }
 
 impl SimpleAsn1Readable<'_> for bool {
@@ -196,6 +288,10 @@ impl SimpleAsn1Writable for bool {
             dest.push_byte(0x00)
         }
     }
+
+    fn data_length(&self) -> Option<usize> {
+        Some(1)
+    }
 }
 
 impl<'a> SimpleAsn1Readable<'a> for &'a [u8] {
@@ -210,6 +306,10 @@ impl SimpleAsn1Writable for &[u8] {
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
         dest.push_slice(self)
     }
+
+    fn data_length(&self) -> Option<usize> {
+        Some(self.len())
+    }
 }
 
 impl<const N: usize> SimpleAsn1Readable<'_> for [u8; N] {
@@ -225,6 +325,10 @@ impl<const N: usize> SimpleAsn1Writable for [u8; N] {
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
         dest.push_slice(self)
     }
+
+    fn data_length(&self) -> Option<usize> {
+        Some(N)
+    }
 }
 
 /// Represents values that are encoded as an `OCTET STRING` containing an
@@ -258,12 +362,16 @@ impl<T: Asn1Writable> SimpleAsn1Writable for OctetStringEncoded<T> {
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
         self.0.write(&mut Writer::new(dest))
     }
+
+    fn data_length(&self) -> Option<usize> {
+        self.0.encoded_length()
+    }
 }
 
 /// Type for use with `Parser.read_element` and `Writer.write_element` for
 /// handling ASN.1 `PrintableString`.  A `PrintableString` contains an `&str`
-/// with only valid characers.
-#[derive(Clone, Debug, PartialEq, Eq)]
+/// with only valid characters.
+#[derive(Clone, Debug, PartialEq, Eq, Hash)]
 pub struct PrintableString<'a>(&'a str);
 
 impl<'a> PrintableString<'a> {
@@ -329,11 +437,15 @@ impl SimpleAsn1Writable for PrintableString<'_> {
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
         dest.push_slice(self.0.as_bytes())
     }
+
+    fn data_length(&self) -> Option<usize> {
+        Some(self.0.len())
+    }
 }
 
 /// Type for use with `Parser.read_element` and `Writer.write_element` for
 /// handling ASN.1 `IA5String`.  An `IA5String` contains an `&str`
-/// with only valid characers.
+/// with only valid characters.
 #[derive(Clone, Debug, PartialEq, Eq)]
 pub struct IA5String<'a>(&'a str);
 
@@ -378,11 +490,15 @@ impl SimpleAsn1Writable for IA5String<'_> {
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
         dest.push_slice(self.0.as_bytes())
     }
+
+    fn data_length(&self) -> Option<usize> {
+        Some(self.0.len())
+    }
 }
 
 /// Type for use with `Parser.read_element` and `Writer.write_element` for
 /// handling ASN.1 `UTF8String`.
-#[derive(Clone, Debug, PartialEq, Eq)]
+#[derive(Clone, Debug, PartialEq, Eq, Hash)]
 pub struct Utf8String<'a>(&'a str);
 
 impl<'a> Utf8String<'a> {
@@ -411,11 +527,15 @@ impl SimpleAsn1Writable for Utf8String<'_> {
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
         dest.push_slice(self.0.as_bytes())
     }
+
+    fn data_length(&self) -> Option<usize> {
+        Some(self.0.len())
+    }
 }
 
 /// Type for use with `Parser.read_element` and `Writer.write_element` for
-/// handling ASN.1 `VisibleString`.  An `VisibleString` contains an `&str`
-/// with only valid characers.
+/// handling ASN.1 `VisibleString`.  A `VisibleString` contains an `&str`
+/// with only valid characters.
 #[derive(Clone, Debug, PartialEq, Eq)]
 pub struct VisibleString<'a>(&'a str);
 
@@ -466,12 +586,16 @@ impl SimpleAsn1Writable for VisibleString<'_> {
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
         dest.push_slice(self.0.as_bytes())
     }
+
+    fn data_length(&self) -> Option<usize> {
+        Some(self.0.len())
+    }
 }
 
 /// Type for use with `Parser.read_element` and `Writer.write_element` for
 /// handling ASN.1 `BMPString`. A `BMPString` contains encoded (UTF-16-BE)
 /// bytes which are known to be valid.
-#[derive(Debug, PartialEq, Clone, Eq)]
+#[derive(Debug, PartialEq, Clone, Eq, Hash)]
 pub struct BMPString<'a>(&'a [u8]);
 
 impl<'a> BMPString<'a> {
@@ -516,12 +640,16 @@ impl SimpleAsn1Writable for BMPString<'_> {
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
         dest.push_slice(self.as_utf16_be_bytes())
     }
+
+    fn data_length(&self) -> Option<usize> {
+        Some(self.as_utf16_be_bytes().len())
+    }
 }
 
 /// Type for use with `Parser.read_element` and `Writer.write_element` for
 /// handling ASN.1 `UniversalString`. A `UniversalString` contains encoded
 /// (UTF-32-BE) bytes which are known to be valid.
-#[derive(Debug, PartialEq, Clone, Eq)]
+#[derive(Debug, PartialEq, Clone, Eq, Hash)]
 pub struct UniversalString<'a>(&'a [u8]);
 
 impl<'a> UniversalString<'a> {
@@ -566,9 +694,13 @@ impl SimpleAsn1Writable for UniversalString<'_> {
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
         dest.push_slice(self.as_utf32_be_bytes())
     }
+
+    fn data_length(&self) -> Option<usize> {
+        Some(self.as_utf32_be_bytes().len())
+    }
 }
 
-fn validate_integer(data: &[u8], signed: bool) -> ParseResult<()> {
+const fn validate_integer(data: &[u8], signed: bool) -> ParseResult<()> {
     if data.is_empty() {
         return Err(ParseError::new(ParseErrorKind::InvalidValue));
     }
@@ -616,6 +748,16 @@ macro_rules! impl_asn1_element_for_int {
         impl SimpleAsn1Writable for $t {
             const TAG: Tag = Tag::primitive(0x02);
             fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
+                let num_bytes = self.data_length().unwrap() as u32;
+
+                for i in (1..=num_bytes).rev() {
+                    let digit = self.checked_shr((i - 1) * 8).unwrap_or(0);
+                    dest.push_byte(digit as u8)?;
+                }
+                Ok(())
+            }
+
+            fn data_length(&self) -> Option<usize> {
                 let mut num_bytes = 1;
                 let mut v: $t = *self;
                 #[allow(unused_comparisons)]
@@ -623,12 +765,22 @@ macro_rules! impl_asn1_element_for_int {
                     num_bytes += 1;
                     v = v.checked_shr(8).unwrap_or(0);
                 }
+                Some(num_bytes)
+            }
+        }
+    };
+}
+macro_rules! impl_asn1_write_for_nonzero_int {
+    ($t:ty, $inner_type:ty) => {
+        impl SimpleAsn1Writable for $t {
+            const TAG: Tag = <$inner_type as SimpleAsn1Writable>::TAG;
 
-                for i in (1..=num_bytes).rev() {
-                    let digit = self.checked_shr((i - 1) * 8).unwrap_or(0);
-                    dest.push_byte(digit as u8)?;
-                }
-                Ok(())
+            fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
+                <$inner_type>::from(*self).write_data(dest)
+            }
+
+            fn data_length(&self) -> Option<usize> {
+                <$inner_type>::from(*self).data_length()
             }
         }
     };
@@ -643,6 +795,15 @@ impl_asn1_element_for_int!(u32; false);
 impl_asn1_element_for_int!(i64; true);
 impl_asn1_element_for_int!(u64; false);
 
+impl_asn1_write_for_nonzero_int!(NonZeroI8, i8);
+impl_asn1_write_for_nonzero_int!(NonZeroI16, i16);
+impl_asn1_write_for_nonzero_int!(NonZeroI32, i32);
+impl_asn1_write_for_nonzero_int!(NonZeroI64, i64);
+impl_asn1_write_for_nonzero_int!(NonZeroU8, u8);
+impl_asn1_write_for_nonzero_int!(NonZeroU16, u16);
+impl_asn1_write_for_nonzero_int!(NonZeroU32, u32);
+impl_asn1_write_for_nonzero_int!(NonZeroU64, u64);
+
 /// Arbitrary sized unsigned integer. Contents may be accessed as `&[u8]` of
 /// big-endian data. Its contents always match the DER encoding of a value
 /// (i.e. they are minimal)
@@ -656,9 +817,11 @@ impl<'a> BigUint<'a> {
     /// as required by DER: minimally and if the high bit would be set in the
     /// first octet, a leading `\x00` should be prepended (to disambiguate from
     /// negative values).
-    pub fn new(data: &'a [u8]) -> Option<Self> {
-        validate_integer(data, false).ok()?;
-        Some(BigUint { data })
+    pub const fn new(data: &'a [u8]) -> Option<Self> {
+        match validate_integer(data, false) {
+            Ok(()) => Some(BigUint { data }),
+            Err(_) => None,
+        }
     }
 
     /// Returns the contents of the integer as big-endian bytes.
@@ -678,6 +841,10 @@ impl SimpleAsn1Writable for BigUint<'_> {
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
         dest.push_slice(self.as_bytes())
     }
+
+    fn data_length(&self) -> Option<usize> {
+        Some(self.as_bytes().len())
+    }
 }
 
 /// Arbitrary sized unsigned integer which owns its data. Contents may be
@@ -716,6 +883,10 @@ impl SimpleAsn1Writable for OwnedBigUint {
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
         dest.push_slice(self.as_bytes())
     }
+
+    fn data_length(&self) -> Option<usize> {
+        Some(self.as_bytes().len())
+    }
 }
 
 /// Arbitrary sized signed integer. Contents may be accessed as `&[u8]` of
@@ -731,9 +902,11 @@ impl<'a> BigInt<'a> {
     /// as required by DER: minimally and if the high bit would be set in the
     /// first octet, a leading `\x00` should be prepended (to disambiguate from
     /// negative values).
-    pub fn new(data: &'a [u8]) -> Option<Self> {
-        validate_integer(data, true).ok()?;
-        Some(BigInt { data })
+    pub const fn new(data: &'a [u8]) -> Option<Self> {
+        match validate_integer(data, true) {
+            Ok(()) => Some(BigInt { data }),
+            Err(_) => None,
+        }
     }
 
     /// Returns the contents of the integer as big-endian bytes.
@@ -758,6 +931,10 @@ impl SimpleAsn1Writable for BigInt<'_> {
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
         dest.push_slice(self.as_bytes())
     }
+
+    fn data_length(&self) -> Option<usize> {
+        Some(self.as_bytes().len())
+    }
 }
 
 /// Arbitrary sized signed integer which owns its contents. Contents may be
@@ -800,6 +977,10 @@ impl SimpleAsn1Writable for OwnedBigInt {
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
         dest.push_slice(self.as_bytes())
     }
+
+    fn data_length(&self) -> Option<usize> {
+        Some(self.as_bytes().len())
+    }
 }
 
 impl<'a> SimpleAsn1Readable<'a> for ObjectIdentifier {
@@ -813,6 +994,9 @@ impl SimpleAsn1Writable for ObjectIdentifier {
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
         dest.push_slice(self.as_der())
     }
+    fn data_length(&self) -> Option<usize> {
+        Some(self.as_der().len())
+    }
 }
 
 impl<'a> SimpleAsn1Readable<'a> for BitString<'a> {
@@ -831,6 +1015,10 @@ impl SimpleAsn1Writable for BitString<'_> {
         dest.push_byte(self.padding_bits())?;
         dest.push_slice(self.as_bytes())
     }
+
+    fn data_length(&self) -> Option<usize> {
+        Some(1 + self.as_bytes().len())
+    }
 }
 impl<'a> SimpleAsn1Readable<'a> for OwnedBitString {
     const TAG: Tag = Tag::primitive(0x03);
@@ -844,6 +1032,10 @@ impl SimpleAsn1Writable for OwnedBitString {
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
         self.as_bitstring().write_data(dest)
     }
+
+    fn data_length(&self) -> Option<usize> {
+        self.as_bitstring().data_length()
+    }
 }
 
 fn read_byte(data: &mut &[u8]) -> ParseResult<u8> {
@@ -874,7 +1066,7 @@ fn read_4_digits(data: &mut &[u8]) -> ParseResult<u16> {
         + u16::from(read_digit(data)?))
 }
 
-fn validate_date(year: u16, month: u8, day: u8) -> ParseResult<()> {
+const fn validate_date(year: u16, month: u8, day: u8) -> ParseResult<()> {
     if day < 1 {
         return Err(ParseError::new(ParseErrorKind::InvalidValue));
     }
@@ -934,7 +1126,7 @@ pub struct DateTime {
 }
 
 impl DateTime {
-    pub fn new(
+    pub const fn new(
         year: u16,
         month: u8,
         day: u8,
@@ -942,18 +1134,20 @@ impl DateTime {
         minute: u8,
         second: u8,
     ) -> ParseResult<DateTime> {
-        validate_date(year, month, day)?;
         if hour > 23 || minute > 59 || second > 59 {
             return Err(ParseError::new(ParseErrorKind::InvalidValue));
         }
-        Ok(DateTime {
-            year,
-            month,
-            day,
-            hour,
-            minute,
-            second,
-        })
+        match validate_date(year, month, day) {
+            Ok(()) => Ok(DateTime {
+                year,
+                month,
+                day,
+                hour,
+                minute,
+                second,
+            }),
+            Err(e) => Err(e),
+        }
     }
 
     /// The calendar year.
@@ -1045,6 +1239,10 @@ impl SimpleAsn1Writable for UtcTime {
 
         dest.push_byte(b'Z')
     }
+
+    fn data_length(&self) -> Option<usize> {
+        Some(13)
+    }
 }
 
 /// Used for parsing and writing ASN.1 `GENERALIZED TIME` values used in X.509.
@@ -1094,6 +1292,10 @@ impl SimpleAsn1Writable for X509GeneralizedTime {
 
         dest.push_byte(b'Z')
     }
+
+    fn data_length(&self) -> Option<usize> {
+        Some(15) // YYYYMMDDHHMMSSZ
+    }
 }
 
 /// Used for parsing and writing ASN.1 `GENERALIZED TIME` values,
@@ -1213,6 +1415,19 @@ impl SimpleAsn1Writable for GeneralizedTime {
 
         dest.push_byte(b'Z')
     }
+
+    fn data_length(&self) -> Option<usize> {
+        let base_len = 15; // YYYYMMDDHHMMSSZ
+        if let Some(nanoseconds) = self.nanoseconds() {
+            let mut buf = itoa::Buffer::new();
+            let nanos = buf.format(nanoseconds);
+            let pad = 9 - nanos.len();
+            let nanos = nanos.trim_end_matches('0');
+            Some(base_len + 1 + pad + nanos.len()) // . + padded nanos
+        } else {
+            Some(base_len)
+        }
+    }
 }
 
 /// An ASN.1 `ENUMERATED` value.
@@ -1243,6 +1458,10 @@ impl SimpleAsn1Writable for Enumerated {
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
         u32::write_data(&self.value(), dest)
     }
+
+    fn data_length(&self) -> Option<usize> {
+        self.value().data_length()
+    }
 }
 
 impl<'a, T: Asn1Readable<'a>> Asn1Readable<'a> for Option<T> {
@@ -1268,17 +1487,27 @@ impl<T: Asn1Writable> Asn1Writable for Option<T> {
             Ok(())
         }
     }
+
+    fn encoded_length(&self) -> Option<usize> {
+        match self {
+            Some(v) => v.encoded_length(),
+            None => Some(0),
+        }
+    }
 }
 
 macro_rules! declare_choice {
-    ($count:ident => $(($number:ident $name:ident)),*) => {
-        /// Represents an ASN.1 `CHOICE` with the provided number of potential types.
+    ($count:ident => $(($number:ident $name:ident)),+) => {
+        /// Represents an ASN.1 `CHOICE` with the provided number of potential
+        /// types.
         ///
-        /// If you need more variants than are provided, please file an issue or submit a pull
-        /// request!
+        /// If you need more variants than are provided, please file an issue
+        /// or submit a pull request! Arbitrary numbers of variants are
+        /// supported by the `#[derive(asn1::Asn1Read)]` and
+        /// `#[derive(asn1::Asn1Write)]` APIs.
         #[derive(Debug, PartialEq, Eq)]
         pub enum $count<
-            $($number,)*
+            $($number,)+
         > {
             $(
                 $name($number),
@@ -1290,14 +1519,14 @@ macro_rules! declare_choice {
             $(
                 $number: Asn1Readable<'a>,
             )*
-        > Asn1Readable<'a> for $count<$($number,)*> {
+        > Asn1Readable<'a> for $count<$($number,)+> {
             fn parse(parser: &mut Parser<'a>) -> ParseResult<Self> {
                 let tlv = parser.read_tlv()?;
                 $(
                     if $number::can_parse(tlv.tag()) {
                         return Ok($count::$name(tlv.parse::<$number>()?));
                     }
-                )*
+                )+
                 Err(ParseError::new(ParseErrorKind::UnexpectedTag{actual: tlv.tag()}))
             }
 
@@ -1306,7 +1535,7 @@ macro_rules! declare_choice {
                     if $number::can_parse(tag) {
                         return true;
                     }
-                )*
+                )+
                 false
             }
         }
@@ -1314,13 +1543,21 @@ macro_rules! declare_choice {
         impl<
             $(
                 $number: Asn1Writable,
-            )*
-        > Asn1Writable for $count<$($number,)*> {
+            )+
+        > Asn1Writable for $count<$($number,)+> {
             fn write(&self, w: &mut Writer<'_>) -> WriteResult {
                 match self {
                     $(
                         $count::$name(v) => w.write_element(v),
-                    )*
+                    )+
+                }
+            }
+
+            fn encoded_length(&self) -> Option<usize> {
+                match self {
+                    $(
+                        $count::$name(v) => Asn1Writable::encoded_length(v),
+                    )+
                 }
             }
         }
@@ -1371,6 +1608,10 @@ impl SimpleAsn1Writable for Sequence<'_> {
     fn write_data(&self, data: &mut WriteBuf) -> WriteResult {
         data.push_slice(self.data)
     }
+
+    fn data_length(&self) -> Option<usize> {
+        Some(self.data.len())
+    }
 }
 
 /// Writes an ASN.1 `SEQUENCE` using a callback that writes the inner
@@ -1392,6 +1633,10 @@ impl SimpleAsn1Writable for SequenceWriter<'_> {
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
         (self.f)(&mut Writer::new(dest))
     }
+
+    fn data_length(&self) -> Option<usize> {
+        None
+    }
 }
 
 /// Represents an ASN.1 `SEQUENCE OF`. This is an `Iterator` over values that
@@ -1537,6 +1782,11 @@ impl<
 
         Ok(())
     }
+
+    fn data_length(&self) -> Option<usize> {
+        let iter = self.clone();
+        iter.map(|el| el.encoded_length()).sum()
+    }
 }
 
 /// Writes a `SEQUENCE OF` ASN.1 structure from a slice of `T`.
@@ -1565,6 +1815,11 @@ impl<T: Asn1Writable, V: Borrow<[T]>> SimpleAsn1Writable for SequenceOfWriter<'_
 
         Ok(())
     }
+
+    fn data_length(&self) -> Option<usize> {
+        let vals = self.vals.borrow();
+        vals.iter().map(|v| v.encoded_length()).sum()
+    }
 }
 
 /// Represents an ASN.1 `SET OF`. This is an `Iterator` over values that
@@ -1677,6 +1932,10 @@ impl<'a, T: Asn1Readable<'a> + Asn1Writable> SimpleAsn1Writable for SetOf<'a, T>
 
         Ok(())
     }
+    fn data_length(&self) -> Option<usize> {
+        let iter = self.clone();
+        iter.map(|el| el.encoded_length()).sum()
+    }
 }
 
 /// Writes an ASN.1 `SET OF` whose contents is a slice of `T`. This type handles
@@ -1729,6 +1988,11 @@ impl<T: Asn1Writable, V: Borrow<[T]>> SimpleAsn1Writable for SetOfWriter<'_, T,
 
         Ok(())
     }
+
+    fn data_length(&self) -> Option<usize> {
+        let vals = self.vals.borrow();
+        vals.iter().map(|v| v.encoded_length()).sum()
+    }
 }
 
 /// `Implicit` is a type which wraps another ASN.1 type, indicating that the tag is an ASN.1
@@ -1773,6 +2037,10 @@ impl<T: SimpleAsn1Writable, const TAG: u32> SimpleAsn1Writable for Implicit<T, {
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
         self.inner.write_data(dest)
     }
+
+    fn data_length(&self) -> Option<usize> {
+        self.inner.data_length()
+    }
 }
 
 /// `Explicit` is a type which wraps another ASN.1 type, indicating that the tag is an ASN.1
@@ -1814,6 +2082,9 @@ impl<T: Asn1Writable, const TAG: u32> SimpleAsn1Writable for Explicit<T, { TAG }
     fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
         Writer::new(dest).write_element(&self.inner)
     }
+    fn data_length(&self) -> Option<usize> {
+        self.inner.encoded_length()
+    }
 }
 
 impl<'a, T: Asn1Readable<'a>, U: Asn1DefinedByReadable<'a, T>, const TAG: u32>
@@ -1834,9 +2105,15 @@ impl<T: Asn1Writable, U: Asn1DefinedByWritable<T>, const TAG: u32> Asn1DefinedBy
         self.as_inner().item()
     }
     fn write(&self, dest: &mut Writer<'_>) -> WriteResult {
-        dest.write_tlv(crate::explicit_tag(TAG), |dest| {
-            self.as_inner().write(&mut Writer::new(dest))
-        })
+        dest.write_tlv(
+            crate::explicit_tag(TAG),
+            self.as_inner().encoded_length(),
+            |dest| self.as_inner().write(&mut Writer::new(dest)),
+        )
+    }
+    fn encoded_length(&self) -> Option<usize> {
+        let inner_len = self.as_inner().encoded_length()?;
+        Some(Tlv::full_length(crate::explicit_tag(TAG), inner_len))
     }
 }
 
@@ -1862,6 +2139,10 @@ impl<T: Asn1Writable> Asn1Writable for DefinedByMarker<T> {
     fn write(&self, _: &mut Writer<'_>) -> WriteResult {
         panic!("write() should never be called on a DefinedByMarker")
     }
+
+    fn encoded_length(&self) -> Option<usize> {
+        panic!("encoded_length() should never be called on a DefinedByMarker")
+    }
 }
 
 #[cfg(test)]
@@ -2233,4 +2514,10 @@ mod tests {
     fn test_defined_by_marker_write() {
         crate::write(|w| DefinedByMarker::<ObjectIdentifier>::marker().write(w)).unwrap();
     }
+
+    #[test]
+    #[should_panic]
+    fn test_defined_by_marker_encoded_length() {
+        DefinedByMarker::<ObjectIdentifier>::marker().encoded_length();
+    }
 }
