@@ -1,4 +1,4 @@
-use crate::types::Asn1Writable;
+use crate::types::{Asn1Writable, SimpleAsn1Writable};
 use crate::Tag;
 #[cfg(not(feature = "std"))]
 use alloc::vec::Vec;
@@ -46,21 +46,26 @@ impl WriteBuf {
         self.0.as_mut_slice()
     }
 
+    // Reserve space for up to `len` additional bytes.
     #[inline]
-    pub fn push_byte(&mut self, b: u8) -> WriteResult {
+    pub fn reserve_additional(&mut self, len: usize) -> WriteResult {
         self.0
-            .try_reserve(1)
+            .try_reserve(len)
             .map_err(|_| WriteError::AllocationError)?;
 
+        Ok(())
+    }
+
+    #[inline]
+    pub fn push_byte(&mut self, b: u8) -> WriteResult {
+        self.reserve_additional(1)?;
         self.0.push(b);
         Ok(())
     }
 
     #[inline]
     pub fn push_slice(&mut self, data: &[u8]) -> WriteResult {
-        self.0
-            .try_reserve(data.len())
-            .map_err(|_| WriteError::AllocationError)?;
+        self.reserve_additional(data.len())?;
 
         self.0.extend_from_slice(data);
         Ok(())
@@ -68,13 +73,17 @@ impl WriteBuf {
 }
 
 fn _length_length(length: usize) -> u8 {
-    let mut i = length;
-    let mut num_bytes = 1;
-    while i > 255 {
-        num_bytes += 1;
-        i >>= 8;
+    (usize::BITS - length.leading_zeros()).div_ceil(8) as u8
+}
+
+/// Calculate the number of bytes needed to encode a length field for the given content length.
+/// This includes the length-of-length byte for lengths >= 128.
+pub(crate) fn length_encoding_size(content_length: usize) -> usize {
+    if content_length < 128 {
+        1
+    } else {
+        1 + _length_length(content_length) as usize
     }
-    num_bytes
 }
 
 fn _insert_at_position(buf: &mut WriteBuf, pos: usize, data: &[u8]) -> WriteResult {
@@ -104,25 +113,53 @@ impl Writer<'_> {
     /// Writes a single element to the output.
     #[inline]
     pub fn write_element<T: Asn1Writable>(&mut self, val: &T) -> WriteResult {
+        if let Some(len) = val.encoded_length() {
+            self.buf.reserve_additional(len)?;
+        }
         val.write(self)
     }
 
     /// Writes a TLV with the specified tag where the value is any bytes
     /// written to the `Vec` in the callback. The length portion of the
     /// TLV is automatically computed.
+    ///
+    /// If `content_length` is provided, it reduces the number of
+    /// re-allocations required.
     #[inline]
     pub fn write_tlv<F: FnOnce(&mut WriteBuf) -> WriteResult>(
         &mut self,
         tag: Tag,
+        content_length: Option<usize>,
         body: F,
     ) -> WriteResult {
         tag.write_bytes(self.buf)?;
-        // Push a 0-byte placeholder for the length. Needing only a single byte
-        // for the element is probably the most common case.
-        self.buf.push_byte(0)?;
-        let start_len = self.buf.len();
-        body(self.buf)?;
-        self.insert_length(start_len)
+
+        match content_length {
+            Some(len) => {
+                // Optimized path: write the correct length encoding upfront
+                if len < 128 {
+                    self.buf.push_byte(len as u8)?;
+                } else {
+                    let num_length_bytes = _length_length(len);
+                    self.buf.push_byte(0x80 | num_length_bytes)?;
+                    for i in (1..=num_length_bytes).rev() {
+                        self.buf.push_byte((len >> ((i - 1) * 8)) as u8)?;
+                    }
+                }
+                let start_len = self.buf.len();
+                body(self.buf)?;
+                assert_eq!(len, self.buf.len() - start_len);
+                Ok(())
+            }
+            None => {
+                // Write a placeholder and then fix the length up later as
+                // required.
+                self.buf.push_byte(0)?;
+                let start_len = self.buf.len();
+                body(self.buf)?;
+                self.insert_length(start_len)
+            }
+        }
     }
 
     #[inline]
@@ -142,6 +179,24 @@ impl Writer<'_> {
 
         Ok(())
     }
+
+    /// This is an alias for `write_element::<Explicit<T, tag>>`.
+    pub fn write_explicit_element<T: Asn1Writable>(&mut self, val: &T, tag: u32) -> WriteResult {
+        let tag = crate::explicit_tag(tag);
+        self.write_tlv(tag, val.encoded_length(), |dest| {
+            Writer::new(dest).write_element(val)
+        })
+    }
+
+    /// This is an alias for `write_element::<Implicit<T, tag>>`.
+    pub fn write_implicit_element<T: SimpleAsn1Writable>(
+        &mut self,
+        val: &T,
+        tag: u32,
+    ) -> WriteResult {
+        let tag = crate::implicit_tag(tag, T::TAG);
+        self.write_tlv(tag, val.data_length(), |dest| val.write_data(dest))
+    }
 }
 
 /// Constructs a writer and invokes a callback which writes ASN.1 elements into
@@ -372,6 +427,94 @@ mod tests {
         ]);
     }
 
+    #[test]
+    fn test_write_nonzeroi8_like_underlying() {
+        use std::num::NonZeroI8;
+        for val in [1, -1, i8::MIN, i8::MAX] {
+            assert_eq!(
+                write_single::<i8>(&val).unwrap(),
+                write_single(&NonZeroI8::new(val).unwrap()).unwrap()
+            );
+        }
+    }
+
+    #[test]
+    fn test_write_nonzeroi16_like_underlying() {
+        use std::num::NonZeroI16;
+        for val in [1, -1, i16::MIN, i16::MAX] {
+            assert_eq!(
+                write_single::<i16>(&val).unwrap(),
+                write_single(&NonZeroI16::new(val).unwrap()).unwrap()
+            );
+        }
+    }
+
+    #[test]
+    fn test_write_nonzeroi32_like_underlying() {
+        use std::num::NonZeroI32;
+        for val in [1, -1, i32::MIN, i32::MAX] {
+            assert_eq!(
+                write_single::<i32>(&val).unwrap(),
+                write_single(&NonZeroI32::new(val).unwrap()).unwrap()
+            );
+        }
+    }
+
+    #[test]
+    fn test_write_nonzeroi64_like_underlying() {
+        use std::num::NonZeroI64;
+        for val in [1, -1, i64::MIN, i64::MAX] {
+            assert_eq!(
+                write_single::<i64>(&val).unwrap(),
+                write_single(&NonZeroI64::new(val).unwrap()).unwrap()
+            );
+        }
+    }
+
+    #[test]
+    fn test_write_nonzerou8_like_underlying() {
+        use std::num::NonZeroU8;
+        for val in [1, u8::MAX] {
+            assert_eq!(
+                write_single::<u8>(&val).unwrap(),
+                write_single(&NonZeroU8::new(val).unwrap()).unwrap()
+            );
+        }
+    }
+
+    #[test]
+    fn test_write_nonzerou16_like_underlying() {
+        use std::num::NonZeroU16;
+        for val in [1, u16::MAX] {
+            assert_eq!(
+                write_single::<u16>(&val).unwrap(),
+                write_single(&NonZeroU16::new(val).unwrap()).unwrap()
+            );
+        }
+    }
+
+    #[test]
+    fn test_write_nonzerou32_like_underlying() {
+        use std::num::NonZeroU32;
+        for val in [1, u32::MAX] {
+            assert_eq!(
+                write_single::<u32>(&val).unwrap(),
+                write_single(&NonZeroU32::new(val).unwrap()).unwrap()
+            );
+        }
+    }
+
+    #[test]
+    fn test_write_nonzerou64_like_underlying() {
+        use std::num::NonZeroU64;
+        for val in [1, u64::MAX] {
+            assert_eq!(
+                write_single::<u64>(&val).unwrap(),
+                write_single(&NonZeroU64::new(val).unwrap()).unwrap()
+            );
+        }
+    }
+
     #[test]
     fn test_write_u8() {
         assert_writes::<u8>(&[
@@ -590,6 +733,14 @@ mod tests {
             .unwrap(),
             b"\x30\x03\x01\x01\xff"
         );
+        assert_eq!(
+            write(|w| {
+                w.write_element(&SequenceWriter::new(&|w: &mut Writer<'_>| {
+                    w.write_element(&b"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa")
+                }))
+            }).unwrap(),
+            b"\x30\x81\x84\x04\x81\x81aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
+        );
 
         assert_writes(&[(
             parse_single::<Sequence<'_>>(b"\x30\x06\x01\x01\xff\x02\x01\x06").unwrap(),
@@ -672,6 +823,17 @@ mod tests {
             (Implicit::new(true), b"\x82\x01\xff"),
             (Implicit::new(false), b"\x82\x01\x00"),
         ]);
+
+        assert_eq!(
+            write(|w| { w.write_implicit_element(&true, 2) }).unwrap(),
+            b"\x82\x01\xff"
+        );
+
+        assert_eq!(
+            write(|w| { w.write_implicit_element(&SequenceWriter::new(&|_w| { Ok(()) }), 2) })
+                .unwrap(),
+            b"\xa2\x00"
+        );
     }
 
     #[test]
@@ -680,6 +842,11 @@ mod tests {
             (Explicit::new(true), b"\xa2\x03\x01\x01\xff"),
             (Explicit::new(false), b"\xa2\x03\x01\x01\x00"),
         ]);
+
+        assert_eq!(
+            write(|w| { w.write_explicit_element(&true, 2) }).unwrap(),
+            b"\xa2\x03\x01\x01\xff"
+        );
     }
 
     #[test]
